package my.com.connector.ws.client.jackson;

import java.io.IOException;
import java.io.InputStream;
import java.sql.Connection;

import javax.servlet.http.HttpServletRequest;

import my.com.connector.ws.client.constants.EventStatus;
import my.com.connector.ws.client.dao.BasicEventLogDao;
import my.com.connector.ws.client.dao.EventLogDao;
import my.com.connector.ws.client.data.AsyncWebServicesRequest;
import my.com.connector.ws.client.data.EventLog;
import my.com.connector.ws.client.data.WebServicesRequest;

import org.apache.commons.io.IOUtils;
import org.codehaus.jackson.map.ObjectMapper;
import org.springframework.context.ApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.client.RestTemplate;

/**
 * A helper class for consuming JSON based web services.
 * 
 * @author chinyeh
 *
 */
public final class WebServicesHelper {

	private static final String APP_CONTEXT_FILENAME = 
		"META-INF/spring/ConnectorClientUtils-context.xml";
	
	private static final RestTemplate REST_TEMPLATE;
	
	private static final ObjectMapper OBJECT_MAPPER;
	
	static {
		ApplicationContext appContext = 
			new ClassPathXmlApplicationContext(APP_CONTEXT_FILENAME);
		
		REST_TEMPLATE = (RestTemplate) appContext.getBean("restTemplate");
		OBJECT_MAPPER = new ObjectMapper();
	}
	
	/**
	 * Helper class.
	 */
	private WebServicesHelper() {
	}
	
	/**
	 * Invoke the web services.
	 * 
	 * @param <E> the type of the request object
	 * @param <T> the type of the return object
	 * @param request the web service request
	 * @return the response of the web service call
	 */
	public static <E, T> T invoke(WebServicesRequest<E, T> request) {
		HttpHeaders httpHeaders = createJsonHttpHeaders();
		HttpEntity<E> httpEntity = 
			new HttpEntity<E>(request.getRequestEntity(), httpHeaders);
		
		ResponseEntity<T> response = REST_TEMPLATE.exchange(
				request.getRequestUrl(), 
				request.getHttpMethod(), 
				httpEntity, 
				request.getResponseType());
		
		return response.getBody();
	}
	
	/**
	 * Persist the event and then invoke the asynchronous based 
	 * web services call.
	 * 
	 * @param <E> the type of the request object
	 * @param <T> the type of the return object
	 * @param connection an reference of the established database connection
	 * @param request the web service request
	 */
	public static <E, T> void persistAndInvoke(Connection connection,
			AsyncWebServicesRequest<E, T> request) {
		
		EventLog eventLog = persistEvent(connection, request);
		request.setEventLogId(Integer.toString(eventLog.getId()));
		
		HttpHeaders httpHeaders = createJsonHttpHeaders();
		addApplicationHeaders(httpHeaders, request);
		
		HttpEntity<E> httpEntity = 
			new HttpEntity<E>(request.getRequestEntity(), httpHeaders);
		
		REST_TEMPLATE.exchange(
				request.getAsyncWebServiceUrl(), 
				request.getHttpMethod(), 
				httpEntity, 
				String.class);
	}
	
	/**
	 * @param connection an reference of an established database connection
	 * @param servletRequest the Http servlet request which contains the 
	 * 		event status
	 */
	public static void updateEventLogStatus(Connection connection,
			HttpServletRequest servletRequest) {
		String eventDetails = "";
		String eventLogId = "";
		String eventStatus = "";
		
		try {
			InputStream rawBodyInputStream = servletRequest.getInputStream();
			
			eventDetails = IOUtils.toString(rawBodyInputStream, "UTF-8");
			eventLogId = servletRequest.getParameter("eventId");
			eventStatus = servletRequest.getParameter("status");
			
			if ((eventLogId != null) && (eventLogId.length() > 0)) {
				updateEvent(connection, eventLogId, eventStatus, eventDetails);
			}
			
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * @param <E> the type of the request entity
	 * @param <T> the type of the response entity
	 * @param httpHeaders the HTTP headers to be added
	 * @param request the web service request
	 */
	private static <E, T> void addApplicationHeaders(HttpHeaders httpHeaders, 
			AsyncWebServicesRequest<E, T> request) {
		httpHeaders.set("xx-eventlog-id-xx", request.getEventLogId());
		httpHeaders.set("xx-forward-url-xx", request.getRequestUrl());
		httpHeaders.set("xx-callback-url-xx", request.getCallbackUrl());
		httpHeaders.set("xx-http-forward-method-xx", 
				request.getHttpMethod().name());
	}

	/**
	 * @return the populated HTTP headers
	 */
	private static HttpHeaders createJsonHttpHeaders() {
		HttpHeaders httpHeaders = new HttpHeaders();
		httpHeaders.setContentType(MediaType.APPLICATION_JSON);

		return httpHeaders;
	}
	
	/**
	 * Persists the request into event log and updates the generated 
	 * event log's ID into the given request object.
	 * 
	 * @param <E> the type of the request entity
	 * @param <T> the type of the response entity
	 * @param connection an reference of the established database connection
	 * @param request the request to be persisted
	 * @return the persisted event log
	 */
	private static <E, T> EventLog persistEvent(
			Connection connection,
			AsyncWebServicesRequest<E, T> request) {
		EventLogDao eventLogDao = new BasicEventLogDao(connection);
		
		EventLog eventLog = toEventLog(request);
		eventLog = eventLogDao.persist(eventLog);
		
		return eventLog;
	}
	
	/**
	 * @param <E> the type of the request entity
	 * @param <T> the type of the response entity
	 * @param request the request to be transformed
	 * @return the transformed event log
	 */
	private static <E, T> EventLog toEventLog(
			AsyncWebServicesRequest<E, T> request) {
		EventLog eventLog = new EventLog();
		eventLog.setRequestUrl(request.getRequestUrl());
		eventLog.setMallIndicator(request.getMallIndicator());
		eventLog.setTrxType(request.getTrxType());
		eventLog.setClientTrxId(request.getClientTrxId());
		eventLog.setStatus(EventStatus.NEW.toString());
		eventLog.setCallbackUrl(request.getCallbackUrl());
		eventLog.setHttpMethod(request.getHttpMethod().name());
		
		String requestBodyInStr = toJsonString(request.getRequestEntity());
		eventLog.setRequestBody(requestBodyInStr);
		
		return eventLog;
	}

	/**
	 * @param <T> the type of the object to be serialized
	 * @param requestObject the object to be serialized
	 * @return the serialized object in String
	 */
	private static <T> String toJsonString(T requestObject) {
		String result = "";
		
		if (OBJECT_MAPPER.canSerialize(requestObject.getClass())) {
			try {
				result = OBJECT_MAPPER.writeValueAsString(requestObject);
				
			} catch (IOException ex) {
				throw new RuntimeException("unable to serialize the object",
						ex);
			}
		} else {
			throw new IllegalArgumentException(requestObject.getClass() 
					+ " cannot be serialized");
		}
		
		return result;
	}

	/**
	 * @param connection the reference of an established database connection
	 * @param eventLogId the ID of the event
	 * @param eventStatus the status of the event
	 * @param eventDetails the details of the event
	 */
	private static void updateEvent(Connection connection, String eventLogId, 
			String eventStatus, String eventDetails) {
		EventLogDao eventLogDao = new BasicEventLogDao(connection);
		eventLogDao.updateEventById(Integer.valueOf(eventLogId), eventStatus, 
				eventDetails);
	}
}
